#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Iterable, Generator, Optional
from .typing import PipelineInput

import numpy as np

from pathlib import Path
from .pipeline import Pipeline

import chocolate
from chocolate.base import SearchAlgorithm
from chocolate import SQLiteConnection

class Optimizer:
    """Pipeline optimizer

    Parameters
    ----------
    pipeline : `Pipeline`
        Pipeline.
    db : `Path`
        Path to iteration database on disk.
    sampler : `str`, optional
        One of "Random", "QuasiRandom", "Bayes", or "CMAES".
        Defaults to "QuasiRandom".
    """

    def __init__(self, pipeline: Pipeline,
                       db: Path,
                       sampler: Optional[str] = "QuasiRandom"):

        self.pipeline = pipeline

        self.db = db
        self.connection_ = SQLiteConnection(f'sqlite:///{self.db}')

        self.sampler = sampler

        try:
            Sampler = getattr(chocolate, sampler)
        except AttributeError as e:
            msg = ('"sampler must be one of "Random", "QuasiRandom", '
                   '"Bayes", or "CMAES".')
            raise ValueError(msg)

        if not issubclass(Sampler, SearchAlgorithm):
            msg = ('"sampler must be one of "Random", "QuasiRandom", '
                   '"Bayes", or "CMAES".')
            raise ValueError(msg)

        self.sampler_ = Sampler(self.connection_,
                                self.pipeline.search_space)

    @property
    def best_params(self) -> dict:
        """Return best parameters so far"""
        status = self.status
        if 'best' not in status:
            msg = 'cannot get parameters as no trial has finished yet.'
            raise AttributeError(msg)
        return status['best']['params']

    @property
    def best_pipeline(self) -> Pipeline:
        """Return pipeline instantiated with best parameters so far"""
        return self.pipeline.with_params(self.best_params)

    @property
    def status(self) -> dict:
        """Return current status of optimizer

        Returns
        -------
        status : `dict`
            ['iterations']['count'] (`int`) total number of iterations
            ['iterations']['done'] (`int`) total number of sucessful iterations
            ['best']['loss'] (`float`) best loss so far
            ['best']['params'] (`dict`) best parameters so far

        Note
        ----
        If no iteration has finished, `status` does not provide a 'best' key.
        """

        iterations = self.connection_.all_results()

        # total number of iterations (including those which failed and those
        # currently running)
        count = len(iterations)

        # total number of successful iterations
        done = 0
        for iteration in iterations:
            loss = iteration.get('_loss')
            if loss is None or not np.isfinite(loss):
                continue
            done += 1

        status = {'iterations': {'count': count, 'done': done}}
        if done < 1:
            return status

        def get_loss(iteration):
            loss = iteration.get('_loss')
            if loss is None:
                return np.inf
            return loss

        best_iteration = min(iterations, key=get_loss)
        del best_iteration['id']
        del best_iteration['_chocolate_id']
        loss = best_iteration.pop('_loss')

        params = {str(name): value
                  for name, value in best_iteration.items()}
        # chocolate stores parameters as a flat dictionary
        # while pipeline expects nested dictionaries
        params = self.pipeline.unpack_params(params)
        status['best'] = {'loss': loss, 'params': params}

        return status

    def tune_iter(self, inputs: Iterable[PipelineInput]) -> Generator[dict, None, None]:
        """Tune pipeline forever

        Parameters
        ----------
        inputs : iterable
            List of inputs processed by the pipeline at each iteration.

        Yields
        ------
        status : dict
            ['iteration'] (`int`) latest iteration index
            ['loss'] (`float`) latest iteration loss
            ['params'] (`dict`) latest iteration parameters
            ['new_best'] (`bool`) whether latest iteration is the best so far
        """

        status = self.status
        best_loss = status.get('best', {'loss': np.inf})['loss']

        while True:

            # get next set of hyper-parameters to try
            token, params = self.sampler_.next()

            # depending on the sampler, params maybe float or np.float
            params = {name: value.item() if hasattr(value, 'item') else value
                      for name, value in params.items()}

            # chocolate stores parameters as a flat dictionary
            # while pipeline expects nested dictionaries
            params = self.pipeline.unpack_params(params)

            # instantiate pipeline with this set of parameters
            self.pipeline.with_params(params)

            # NOTE this is embarrasingly parallel. do something about this
            losses = []
            for input in inputs:
                output = self.pipeline(input)
                loss = self.pipeline.loss(input, output)
                losses.append(loss)
            loss = sum(losses) / len(losses)

            result = {'iteration': token['_chocolate_id'] + 1,
                      'loss': loss,
                      'params': params,
                      'new_best': False}

            if loss is not None and loss < best_loss:

                # update best loss from disk as another process might have
                # already reached a better loss in the meantime
                best_loss = self.status.get('best', {'loss': np.inf})['loss']

                if loss < best_loss:
                    result['new_best'] = True
                    best_loss = loss

            self.sampler_.update(token, loss)
            yield result


    def tune(self, inputs: Iterable[PipelineInput],
                   n_iterations: int=10) -> dict:
        """Tune pipeline

        Parameters
        ----------
        inputs : iterable
            List of inputs processed by the pipeline at each iteration.
        n_iterations : int, optional
            Number of iterations. Defaults to 10.

        Returns
        -------
        status : `dict`
            ['iterations']['count'] (`int`) total number of iterations
            ['iterations']['done'] (`int`) total number of sucessful iterations
            ['best']['loss'] (`float`) best loss
            ['best']['params'] (`dict`) best parameters

        Note
        ----
        If no iteration has completed successfully, `status` does not contain
        a 'best' key.
        """

        iterations = self.tune_iter(inputs)

        for i in range(n_iterations):
            _ = next(iterations)

        return self.status



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Optional
from typing import TextIO
from pathlib import Path
from collections import OrderedDict
import chocolate
from .typing import PipelineInput
from .typing import PipelineOutput

from filelock import FileLock
import yaml

from pyannote.core import Timeline
from pyannote.core import Annotation

class Pipeline:
    """Base pipeline"""

    def __init__(self):
        self._hyper_parameters = OrderedDict()
        self._frozen = OrderedDict()
        self._pipelines = OrderedDict()

    def __getattr__(self, name):

        if '_hyper_parameters' in self.__dict__:
            _hyper_parameters = self.__dict__['_hyper_parameters']
            if name in _hyper_parameters:
                return _hyper_parameters[name]

        if '_frozen' in self.__dict__:
            _frozen = self.__dict__['_frozen']
            if name in _frozen:
                return _frozen[name]

        if '_pipelines' in self.__dict__:
            _pipelines = self.__dict__['_pipelines']
            if name in _pipelines:
                return _pipelines[name]

        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, name))

    def __setattr__(self, name, value):

        def remove_from(*dicts):
            for d in dicts:
                if name in d:
                    del d[name]

        _hyper_parameters = self.__dict__.get('_hyper_parameters')
        _frozen = self.__dict__.get('_frozen')
        _pipelines = self.__dict__.get('_pipelines')

        # add/update one hyper-parameter
        if isinstance(value, chocolate.Distribution):
            if _hyper_parameters is None:
                raise AttributeError(
                    "cannot assign hyper-parameters before Pipeline.__init__() call")
            remove_from(self.__dict__, _frozen, _pipelines)
            _hyper_parameters[name] = value
            return

        # add/update one sub-pipeline
        if isinstance(value, Pipeline):
            if _pipelines is None:
                raise AttributeError(
                    "cannot assign pipelines before Pipeline.__init__() call")
            remove_from(self.__dict__, _hyper_parameters, _frozen)
            _pipelines[name] = value
            return

        # freeze hyper-parameter
        if _hyper_parameters is not None and name in _hyper_parameters:
            remove_from(_hyper_parameters)
            _frozen[name] = value
            return

        object.__setattr__(self, name, value)

    def __delattr__(self, name):

        if name in self._hyper_parameters:
            del self._hyper_parameters[name]

        elif name in self._pipelines:
            del self._pipelines[name]

        elif name in self._frozen:
            del self._frozen[name]

        else:
            object.__delattr__(self, name)

    @property
    def search_space(self):
        """Search space (used by pyannote.pipeline.Optimizer)"""

        space = OrderedDict()

        for sub_pipeline_name, sub_pipeline in self._pipelines.items():
            for param_name, param_value in sub_pipeline.search_space.items():
                space[f'{sub_pipeline_name}>{param_name}'] = param_value

        for param_name, param_value in self._hyper_parameters.items():
            space[param_name] = param_value

        return space

    @property
    def params(self):
        """Returns current set of parameters"""
        params = dict(self._hyper_parameters)
        params.update(self._frozen)
        for sub_pipeline_name, sub_pipeline in self._pipelines.items():
            params[sub_pipeline_name] = dict()
            for param_name, param_value in sub_pipeline.params.items():
                params[sub_pipeline_name][param_name] = param_value
        return params

    def instantiate(self):
        """Instantiate root pipeline with current set of hyper-parameters"""
        pass

    def freeze(self, params: dict) -> 'Pipeline':
        """Freeze pipeline parameters

        Parameters
        ----------
        params : `dict`
            Parameters.

        Returns
        -------
        self : `Pipeline`
            Pipeline with frozen parameters
        """

        _hyper_parameters = self.__dict__.get('_hyper_parameters')
        if _hyper_parameters is None:
            msg = "cannot freeze hyper-parameters before they are defined"
            raise AttributeError(msg)

        for name, value in params.items():

            # freeze sub-pipeline hyper-parameter
            if name in self._pipelines:

                if not isinstance(value, dict):
                    msg = (f"only hyper-parameters of '{name}' pipeline can "
                           f"be frozen (not the whole pipeline)")
                    raise ValueError(msg)

                self._pipelines[name].freeze(value)
                continue

            # freeze (or update frozen) root hyper-parameter
            if name in self._hyper_parameters or name in self._frozen:
                setattr(self, name, value)
                continue

            msg = f"hyper-parameter '{name}' does not exist"
            raise ValueError(msg)

        return self

    def with_params(self, params: dict) -> 'Pipeline':
        """Recursively instantiate all pipelines

        Parameters
        ----------
        params : `dict`
            Parameters

        Returns
        -------
        self : `Pipeline`
            Instantiated pipeline.
        """

        for name, value in params.items():

            # subpipeline
            if name in self._pipelines:
                if not isinstance(value, dict):
                    msg = (f"only hyper-parameters of '{name}' pipeline can "
                           f"be instantiated (not the whole pipeline)")
                    raise ValueError(msg)
                self._pipelines[name].with_params(value)
                continue

            # frozen hyper-parameters
            if name in self._frozen:
                if value != self._frozen[name]:
                    msg = (f"cannot change value of frozen hyper-parameter "
                           f"'{name}'")
                    raise ValueError(msg)
                continue

            # regular hyper-parameter
            if name in self._hyper_parameters:
                self._hyper_parameters[name] = value
                continue

            # anything else
            msg = f"hyper-parameter '{name}' does not exist"
            raise ValueError(msg)

        self.instantiate()

        return self

    def __call__(self, input: PipelineInput) -> PipelineOutput:
        """Apply pipeline on input and return its output"""
        raise NotImplementedError

    def loss(self, input: PipelineInput,
                   output: PipelineOutput) -> float:
        """Compute loss for given input/output pair"""
        raise NotImplementedError

    def write(self, file: TextIO,
                    output: PipelineOutput):
        """Write pipeline output to file"""

        if isinstance(output, Timeline):
            for s in output:
                file.write(f'{output.uri} {s.start:.3f} {s.end:.3f}\n')
            return

        if isinstance(output, Annotation):
            for s, t, l in output.itertracks(yield_label=True):
                file.write(f'{output.uri} {output.modality} {s.start:.3f} {s.end:.3f} {t} {l}\n')
            return

        raise NotImplementedError

    def unpack_params(self, params: dict) -> dict:
        """Unpack parameter dictionary

        param1                   param1
        param2                   param2
        sub_pipeline      <--
            sub_param1           sub_pipeline__sub_param1
            sub_param2           sub_pipeline__sub_param2

        Parameters
        ----------
        params : `dict`
            Flat dictionary of parameters.

        Returns
        -------
        unpacked_params : `dict`
            Nested dictionaries of parameters.
        """

        unpacked_params = {}

        pipeline_params = {name: {} for name in self._pipelines}
        for name, value in params.items():
            tokens = name.split('>')
            if len(tokens) > 1:
                pipeline_name = tokens[0]
                param_name = '>'.join(tokens[1:])
                pipeline_params[pipeline_name][param_name] = value
            else:
                unpacked_params[name] = value

        for name, pipeline in self._pipelines.items():
            unpacked_params[name] = pipeline.unpack_params(pipeline_params[name])

        return unpacked_params

    def dump_params(self, params_yml: Path,
                          params: Optional[dict] = None) -> str:
        """Dump parameters to disk

        Parameters
        ----------
        params_yml : `Path`
            Path to YAML file.
        params : `dict`, optional
            Parameters. Defaults to pipeline current parameters.

        Returns
        -------
        content : `str`
            Content written in `param_yml`.
        """

        if params is None:
            params = self.params
        content = yaml.dump(params, default_flow_style=False)
        with FileLock(params_yml.with_suffix('.lock')):
            with open(params_yml, mode='w') as fp:
                fp.write(content)
        return content

    def load_params(self, params_yml: Path) -> 'Pipeline':
        """Instantiate pipeline using parameters from disk

        Parameters
        ----------
        param_yml : `Path`
            Path to YAML file.

        Returns
        -------
        self : `Pipeline`
            Instantiated pipeline

        """

        with open(params_yml, mode='r') as fp:
            params = yaml.load(fp)
        return self.with_params(params)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import warnings
import numpy as np
from typing import Optional

from scipy.cluster.hierarchy import fcluster
from pyannote.core.utils.hierarchy import linkage

import sklearn.cluster
from scipy.spatial.distance import squareform
from pyannote.core.utils.distance import pdist
from pyannote.core.utils.distance import dist_range
from pyannote.core.utils.distance import l2_normalize

import chocolate
from ..pipeline import Pipeline


class HierarchicalAgglomerativeClustering(Pipeline):
    """Hierarchical agglomerative clustering

    Parameters
    ----------
    method : `str`, optional
        Linkage method. Defaults to 'pool'.
    metric : `str`, optional
        Distance metric. Defaults to 'cosine'
    normalize : `bool`, optional
        L2 normalize vectors before clustering.

    Hyper-parameters
    ----------------
    threshold : `float`
        Do not merge clusters with distance greater than `threshold`.
    """

    def __init__(self, method: Optional[str] = 'pool',
                       metric: Optional[str] = 'cosine',
                       normalize: Optional[bool] = False):

        super().__init__()
        self.method = method
        self.metric = metric
        self.normalize = normalize

        min_dist, max_dist = dist_range(metric=self.metric,
                                        normalize=self.normalize)
        if not np.isfinite(max_dist):
            # this is arbitray and might lead to suboptimal results
            max_dist = 1e6
            msg = (f'bounding distance threshold to {max_dist:g}: '
                   f'this might lead to suboptimal results.')
            warnings.warn(msg)
        self.threshold = chocolate.uniform(min_dist, max_dist)

    def __call__(self, X: np.ndarray) -> np.ndarray:
        """Apply hierarchical agglomerative clustering

        Parameters
        ----------
        X : `np.ndarray`
            (n_samples, n_dimensions) feature vectors.

        Returns
        -------
        y : `np.ndarray`
            (n_samples, ) cluster assignment (between 1 and n_clusters).
        """

        if self.normalize:
            X = l2_normalize(X)

        # compute agglomerative clustering all the way up to one cluster
        Z = linkage(X, method=self.method, metric=self.metric)

        # obtain flat clusters by applying distance threshold
        clusters = fcluster(Z, self.threshold, criterion='distance')

        return clusters


class AffinityPropagationClustering(Pipeline):
    """Clustering based on affinity propagation

    Parameters
    ----------
    metric : `str`, optional
        Distance metric. Defaults to 'cosine'

    Hyper-parameters
    ----------------
    damping : `float`
    preference : `float`
        See `sklearn.cluster.AffinityPropagation`
    """

    def __init__(self, metric: Optional[str] = 'cosine'):

        super().__init__()
        self.metric = metric

        self.damping = chocolate.uniform(0.5, 1.0)
        self.preference = chocolate.uniform(-10., 0.)

    def instantiate(self):
        """Instantiate internal sklearn.cluster.AffinityPropagation"""

        self.affinity_propagation_ = sklearn.cluster.AffinityPropagation(
            damping=self.damping, preference=self.preference,
            affinity='precomputed', max_iter=200, convergence_iter=50)

    def __call__(self, X: np.ndarray) -> np.ndarray:
        """Apply clustering based on affinity propagation

        Parameters
        ----------
        X : `np.ndarray`
            (n_samples, n_dimensions) feature vectors.

        Returns
        -------
        y : `np.ndarray`
            (n_samples, ) cluster assignment (between 1 and n_clusters).
        """

        n_samples, _ = X.shape

        try:
            affinity = -squareform(pdist(X, metric=self.metric))
            clusters = self.affinity_propagation_.fit_predict(affinity)
        except MemoryError as e:
            clusters = np.arange(n_samples)

        if np.any(clusters < 0):
            clusters = np.arange(n_samples)
        clusters += 1

        return clusters



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
Pipeline

Usage:
  pyannote-pipeline train [options] [(--forever | --iterations=<iterations>)] <experiment_dir> <database.task.protocol>
  pyannote-pipeline best [options] <experiment_dir> <database.task.protocol>
  pyannote-pipeline apply [options] <params.yml> <database.task.protocol> <output_dir>
  pyannote-pipeline -h | --help
  pyannote-pipeline --version

Common options:
  <database.task.protocol>   Experimental protocol (e.g. "Etape.SpeakerDiarization.TV")
  --database=<db.yml>        Path to database configuration file.
                             [default: ~/.pyannote/db.yml]
  --subset=<subset>          Set subset. Defaults to 'development' in "train"
                             mode, and to all subsets in "apply" mode.
"train" mode:
  <experiment_dir>           Set experiment root directory. This script expects
                             a configuration file called "config.yml" to live
                             in this directory. See "Configuration file"
                             section below for more details.
  --iterations=<iterations>  Number of iterations. [default: 1]
  --forever                  Iterate forever.
  --sampler=<sampler>        Choose sampler between Random, QuasiRandom, CMAES,
                             or Bayes [default: QuasiRandom].

"apply" mode:
  <params.yml>               Path to hyper-parameters.
  <output_dir>               Directory where to store pipeline output.

Configuration file:
    The configuration of each experiment is described in a file called
    <experiment_dir>/config.yml that describes the pipeline.

    ................... <experiment_dir>/config.yml ...................
    pipeline:
       name: Yin2018
       params:
          sad: tutorials/pipeline/sad
          scd: tutorials/pipeline/scd
          emb: tutorials/pipeline/emb
          metric: angular

    # preprocessors can be used to automatically add keys into
    # each (dict) file obtained from pyannote.database protocols.
    preprocessors:
       audio: ~/.pyannote/db.yml   # load template from YAML file
       video: ~/videos/{uri}.mp4   # define template directly

    # one can freeze some hyper-parameters if needed (e.g. when
    # only part of the pipeline needs to be updated)
    freeze:
       speech_turn_segmentation:
          speech_activity_detection:
              onset: 0.5
              offset: 0.5
    ...................................................................

"train" mode:
    Tune the pipeline hyper-parameters
        <experiment_dir>/<database.task.protocol>.<subset>.yml

"best" mode:
    Display current best loss and corresponding hyper-paramters.

"apply" mode
    Apply the pipeline (with best set of hyper-parameters)

"""

import os
import yaml
import numpy as np
from typing import Optional
from pathlib import Path
from docopt import docopt

import itertools
from tqdm import tqdm
from tensorboardX import SummaryWriter

from pyannote.database import FileFinder
from pyannote.database import get_protocol

from pyannote.core.utils.helper import get_class_by_name
from .optimizer import Optimizer


class Experiment:
    """Pipeline experiment

    Parameters
    ----------
    experiment_dir : `Path`
        Experiment root directory.
    training : `bool`, optional
        Switch to training mode
    """

    CONFIG_YML = '{experiment_dir}/config.yml'
    TRAIN_DIR = '{experiment_dir}/train/{protocol}.{subset}'

    @classmethod
    def from_params_yml(cls, params_yml: Path,
                             training : bool = False) -> 'Experiment':
        """Load pipeline from 'params.yml' file

        Parameters
        ----------
        params_yml : `Path`
            Path to 'params.yml' file.
        training : `bool`, optional
            Switch to training mode.

        Returns
        -------
        xp : `Experiment`
            Pipeline experiment.
        """

        experiment_dir = params_yml.parents[2]
        xp = cls(experiment_dir, training=training)
        xp.pipeline_.load_params(params_yml)
        return xp

    def __init__(self, experiment_dir: Path, training: bool = False):

        super().__init__()

        self.experiment_dir = experiment_dir

        # load configuration file
        config_yml = self.CONFIG_YML.format(experiment_dir=self.experiment_dir)
        with open(config_yml, 'r') as fp:
            self.config_ = yaml.load(fp)

        # initialize preprocessors
        preprocessors = {}
        for key, db_yml in self.config_.get('preprocessors', {}).items():
            try:
                preprocessors[key] = FileFinder(db_yml)
            except FileNotFoundError as e:
                template = db_yml
                preprocessors[key] = template
        self.preprocessors_ = preprocessors

        # initialize pipeline
        pipeline_name = self.config_['pipeline']['name']
        Klass = get_class_by_name(
            pipeline_name, default_module_name='pyannote.audio.pipeline')
        self.pipeline_ = Klass(**self.config_['pipeline'].get('params', {}))

        # freeze  parameters
        if 'freeze' in self.config_:
            params = self.config_['freeze']
            self.pipeline_.freeze(params)

    def train(self, protocol_name: str,
                    subset: str = 'development',
                    n_iterations: int = 1,
                    sampler: str = 'QuasiRandom'):
        """Train pipeline

        Parameters
        ----------
        protocol_name : `str`
            Name of pyannote.database protocol to use.
        subset : `str`, optional
            Use this subset for training. Defaults to 'development'.
        n_iterations : `int`, optional
            Number of iterations. Defaults to 1.
        sampler : `str`, optional
            Choose sampler between Random, QuasiRandom, CMAES, or Bayes.
            Defaults to 'QuasiRandom'.
        """
        train_dir = Path(self.TRAIN_DIR.format(
            experiment_dir=self.experiment_dir,
            protocol=protocol_name,
            subset=subset))
        train_dir.mkdir(parents=True, exist_ok=True)

        protocol = get_protocol(protocol_name, progress=False,
                                preprocessors=self.preprocessors_)

        optimizer = Optimizer(self.pipeline_,
                              train_dir / 'iterations.db',
                              sampler=sampler)

        params_yml = train_dir / 'params.yml'

        # TODO. find a way to remove this "pid" hack
        log_dir = train_dir / str(os.getpid())
        writer = SummaryWriter(log_dir=log_dir)

        progress_bar = tqdm(unit='iteration')
        progress_bar.set_description('Waiting for first iteration to complete')
        progress_bar.update(0)

        inputs = list(getattr(protocol, subset)())
        iterations = optimizer.tune_iter(inputs)

        best_loss = optimizer.status.get('best', {'loss': np.inf})['loss']
        count = itertools.count() if n_iterations < 0 else range(n_iterations)

        for i, status in zip(count, iterations):

            iteration = status['iteration']
            loss = status['loss']
            params = status['params']
            new_best = status['new_best']

            if status['new_best']:
                best_loss = loss
                writer.add_scalar(f'train/{protocol_name}.{subset}/loss',
                                  loss, global_step=iteration)
                self.pipeline_.dump_params(params_yml, params=params)

            # progress bar
            desc = (f'Best = {100 * best_loss:g}% | '
                    f'#{iteration} = {100 * loss:g}%')
            progress_bar.set_description(desc=desc)
            progress_bar.update(1)

    def best(self, protocol_name: str,
                   subset: str = 'development'):
        """Print current best pipeline

        Parameters
        ----------
        protocol_name : `str`
            Name of pyannote.database protocol used for training.
        subset : `str`, optional
            Subset used for training. Defaults to 'development'.
        """

        train_dir = Path(self.TRAIN_DIR.format(
            experiment_dir=self.experiment_dir,
            protocol=protocol_name,
            subset=subset))

        optimizer = Optimizer(self.pipeline_,
                              train_dir / 'iterations.db')

        status = optimizer.status
        count = status['iterations']['count']
        done = status['iterations']['done']

        if 'best' not in status:
            print('Still waiting for at least one iteration to succeed.')
            return

        print(
            f'Loss = {100 * status["best"]["loss"]:g}% '
            f'with the following hyper-parameters:')

        content = yaml.dump(status['best']['params'],
                            default_flow_style=False)
        print(content)

    def apply(self, protocol_name: str,
                    output_dir: Path,
                    subset: Optional[str] = None):
        """Apply current best pipeline

        Parameters
        ----------
        protocol_name : `str`
            Name of pyannote.database protocol to process.
        subset : `str`, optional
            Subset to process. Defaults processing all subsets.
        """

        # file generator
        protocol = get_protocol(protocol_name, progress=True,
                                preprocessors=self.preprocessors_)

        output_dir.mkdir(parents=True, exist_ok=False)
        if subset is None:
            path = output_dir / f'{protocol_name}.all.txt'
        else:
            path = output_dir / f'{protocol_name}.{subset}.txt'

        with open(path, mode='w') as fp:

            if subset is None:
                files = FileFinder.protocol_file_iter(protocol)
            else:
                files = getattr(protocol, subset)()

            for current_file in files:
                output = self.pipeline_(current_file)
                self.pipeline_.write(fp, output)

def main():

    arguments = docopt(__doc__, version='Tunable pipelines')

    protocol_name = arguments['<database.task.protocol>']
    subset = arguments['--subset']

    if arguments['train']:

        if subset is None:
            subset = 'development'

        if arguments['--forever']:
            iterations = -1
        else:
            iterations = int(arguments['--iterations'])

        sampler = arguments['--sampler']

        experiment_dir = Path(arguments['<experiment_dir>'])
        experiment_dir = experiment_dir.expanduser().resolve(strict=True)

        experiment = Experiment(experiment_dir, training=True)
        experiment.train(protocol_name, subset=subset, n_iterations=iterations,
                         sampler=sampler)

    if arguments['best']:

        if subset is None:
            subset = 'development'

        experiment_dir = Path(arguments['<experiment_dir>'])
        experiment_dir = experiment_dir.expanduser().resolve(strict=True)

        experiment = Experiment(experiment_dir, training=False)
        experiment.best(protocol_name, subset=subset)

    if arguments['apply']:

        params_yml = Path(arguments['<params.yml>'])
        params_yml = params_yml.expanduser().resolve(strict=True)

        output_dir = Path(arguments['<output_dir>'])
        output_dir = output_dir.expanduser().resolve(strict=False)

        experiment = Experiment.from_params_yml(params_yml, training=False)
        experiment.apply(protocol_name, output_dir, subset=subset)
